from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_08b_daily import SmpmUlDeviceWaterMeter08BDailyData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_water_meter_08b_daily() -> None:
    case_serialized = bytes.fromhex("68e4eab106a33a78")
    assert SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=112323.3, direct_flow_volume_day_ago=3.5, event_battery_or_temperature_limits=True, temperature=23, event_magnet=False, event_continues_consumption=False, event_case_was_opened=False, event_system_error=False, event_reset=True, event_sensor_error=True, event_no_resource=True, event_flow_speed_is_over_limit=True, event_flow_reverse=False) == SmpmUlDeviceWaterMeter08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyData.serialize(SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=112323.3, direct_flow_volume_day_ago=3.5, event_battery_or_temperature_limits=True, temperature=23, event_magnet=False, event_continues_consumption=False, event_case_was_opened=False, event_system_error=False, event_reset=True, event_sensor_error=True, event_no_resource=True, event_flow_speed_is_over_limit=True, event_flow_reverse=False))  # noqa: E501
    case_serialized = bytes.fromhex("680000000080ffff")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=0.0, event_battery_or_temperature_limits=True, temperature=92, event_magnet=True, event_continues_consumption=True, event_case_was_opened=True, event_system_error=True, event_reset=True, event_sensor_error=True, event_no_resource=True, event_flow_speed_is_over_limit=True, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyData.serialize(SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=0.0, event_battery_or_temperature_limits=True, event_case_was_opened=True, event_continues_consumption=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_magnet=True, event_no_resource=True, event_reset=True, event_sensor_error=True, event_system_error=True, temperature=92))  # noqa: E501
    case_serialized = bytes.fromhex("68000000008080ff")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=0.0, event_battery_or_temperature_limits=True, temperature=-35, event_magnet=True, event_continues_consumption=True, event_case_was_opened=True, event_system_error=True, event_reset=True, event_sensor_error=True, event_no_resource=True, event_flow_speed_is_over_limit=True, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyData.serialize(SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=0.0, event_battery_or_temperature_limits=True, event_case_was_opened=True, event_continues_consumption=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_magnet=True, event_no_resource=True, event_reset=True, event_sensor_error=True, event_system_error=True, temperature=-35))  # noqa: E501
    case_serialized = bytes.fromhex("6800000000ffffff")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=12.7, event_battery_or_temperature_limits=True, temperature=92, event_magnet=True, event_continues_consumption=True, event_case_was_opened=True, event_system_error=True, event_reset=True, event_sensor_error=True, event_no_resource=True, event_flow_speed_is_over_limit=True, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyData.serialize(SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=12.7, event_battery_or_temperature_limits=True, event_case_was_opened=True, event_continues_consumption=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_magnet=True, event_no_resource=True, event_reset=True, event_sensor_error=True, event_system_error=True, temperature=92))  # noqa: E501
    case_serialized = bytes.fromhex("6800000000ff80ff")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=12.7, event_battery_or_temperature_limits=True, temperature=-35, event_magnet=True, event_continues_consumption=True, event_case_was_opened=True, event_system_error=True, event_reset=True, event_sensor_error=True, event_no_resource=True, event_flow_speed_is_over_limit=True, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyData.serialize(SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=12.7, event_battery_or_temperature_limits=True, event_case_was_opened=True, event_continues_consumption=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_magnet=True, event_no_resource=True, event_reset=True, event_sensor_error=True, event_system_error=True, temperature=-35))  # noqa: E501
    case_serialized = bytes.fromhex("680000000080ffff")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=0.0, event_battery_or_temperature_limits=True, temperature=92, event_magnet=True, event_continues_consumption=True, event_case_was_opened=True, event_system_error=True, event_reset=True, event_sensor_error=True, event_no_resource=True, event_flow_speed_is_over_limit=True, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyData.serialize(SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=8589934.592, direct_flow_volume_day_ago=0.0, event_battery_or_temperature_limits=True, event_case_was_opened=True, event_continues_consumption=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_magnet=True, event_no_resource=True, event_reset=True, event_sensor_error=True, event_system_error=True, temperature=92))  # noqa: E501
    case_serialized = bytes.fromhex("68000000008080ff")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=0.0, event_battery_or_temperature_limits=True, temperature=-35, event_magnet=True, event_continues_consumption=True, event_case_was_opened=True, event_system_error=True, event_reset=True, event_sensor_error=True, event_no_resource=True, event_flow_speed_is_over_limit=True, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyData.serialize(SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=8589934.592, direct_flow_volume_day_ago=0.0, event_battery_or_temperature_limits=True, event_case_was_opened=True, event_continues_consumption=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_magnet=True, event_no_resource=True, event_reset=True, event_sensor_error=True, event_system_error=True, temperature=-35))  # noqa: E501
    case_serialized = bytes.fromhex("6800000000ffffff")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=12.7, event_battery_or_temperature_limits=True, temperature=92, event_magnet=True, event_continues_consumption=True, event_case_was_opened=True, event_system_error=True, event_reset=True, event_sensor_error=True, event_no_resource=True, event_flow_speed_is_over_limit=True, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyData.serialize(SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=8589934.592, direct_flow_volume_day_ago=12.7, event_battery_or_temperature_limits=True, event_case_was_opened=True, event_continues_consumption=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_magnet=True, event_no_resource=True, event_reset=True, event_sensor_error=True, event_system_error=True, temperature=92))  # noqa: E501
    case_serialized = bytes.fromhex("6800000000ff80ff")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=0.0, direct_flow_volume_day_ago=12.7, event_battery_or_temperature_limits=True, temperature=-35, event_magnet=True, event_continues_consumption=True, event_case_was_opened=True, event_system_error=True, event_reset=True, event_sensor_error=True, event_no_resource=True, event_flow_speed_is_over_limit=True, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyData.serialize(SmpmUlDeviceWaterMeter08BDailyData(direct_flow_volume=8589934.592, direct_flow_volume_day_ago=12.7, event_battery_or_temperature_limits=True, event_case_was_opened=True, event_continues_consumption=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_magnet=True, event_no_resource=True, event_reset=True, event_sensor_error=True, event_system_error=True, temperature=-35))  # noqa: E501



import base64
import json
from datetime import datetime, timedelta, timezone
from enum import Enum, unique
from itertools import groupby
from typing import Callable, Optional, List, Dict, Tuple, Set, Union, NamedTuple, Mapping, Any
from uuid import UUID
from zoneinfo import ZoneInfo

import pytz
from data_aggregator_sdk.constants.enums import NetworkTypeEnum, NetworkSysTypeEnum, EncryptionType as DaEncryptionType, \
    IntegrationV0MessageErrorType, DeviceHack
from data_aggregator_sdk.data_aggregator_api_sdk import DeviceDataOptimized as DADeviceData, ApiDataGatewayNetworkDevice, \
    DataAggregatorApiSdk, ApiDataGatewaysNetworkResponsekOptimized, ApiDataGatewayNetworkOptimized, ApiDataGatewayNetworkResponse
from data_aggregator_sdk.errors import DataAggregatorRequestError, DataAggregatorResponseError
from data_aggregator_sdk.integration_message import IntegrationV0MessageMeta, IntegrationV0MessageGateway, \
    IntegrationV0MessageData, IntegrationV0MessageError
from data_aggregator_sdk.types.get_data_gateway_network_device_list import ApiDataGatewaysNetworkResponse
from pydantic import BaseModel
from timezonefinder import TimezoneFinder
from ul_api_utils.utils.json_encoder import CustomJSONEncoder
from ul_unipipeline.message.uni_message import UniMessage
from ul_unipipeline.worker.uni_worker import UniWorker

from data_gateway_sdk.device_data_encryption import EncryptionType
from data_gateway_sdk.device_data_protocol import DeviceProtocolType
from data_gateway_sdk.errors import DataGatewayDeviceProtocolParsingError
from data_gateway_sdk.protocols.device_packet.device_packet import DevicePacket
from data_gateway_sdk.protocols.nero_bs_packet.http_nero_bs_packet import BaseStationInfoGeo
from src.conf.self_logging import self_logging
from src.data_gateway__memory_db.repositories.network_repository import network_redis_manager

logger = self_logging.getLogger(__name__)

_timezone_finder = TimezoneFinder()


class DataGatewayNetwork(NamedTuple):
    id: UUID
    name: str
    type_network: NetworkTypeEnum
    sys_type: NetworkSysTypeEnum
    specifier: Optional[str]
    params: Optional[Dict[str, Any]]


class OutputMessage(UniMessage):
    data: IntegrationV0MessageGateway
    specifier: str
    network_id: Optional[UUID] = None
    params: Dict[str, Any]


class OutputListMessage(UniMessage):
    messages: List[OutputMessage]


class DeviceDataStructSendingRecipe(NamedTuple):
    net_sys_type: NetworkSysTypeEnum
    network: Optional[DataGatewayNetwork | ApiDataGatewaysNetworkResponse | ApiDataGatewaysNetworkResponsekOptimized]
    message: IntegrationV0MessageGateway


class PayloadInfo(BaseModel):
    encrypted_payload: Optional[str] = ''
    decrypted_payload: Optional[str] = ''
    encryption_type: Optional[EncryptionType] = None


ALL_SUPPORTED_OUTPUT_NETWORK_TYPES = {
    NetworkSysTypeEnum.OUTPUT_REDIRECT_NETWORK,
    NetworkSysTypeEnum.OUTPUT_DATA_AGGREGATOR_DEVICE_DATA,
    NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA,
}


ALL_SUPPORTED_OUTPUT_NETWORK_TYPES_BATCH = {
    (NetworkSysTypeEnum.OUTPUT_REDIRECT_NETWORK, ''),
    (NetworkSysTypeEnum.OUTPUT_REDIRECT_NETWORK, 'repeater_to_dev__batch'),
    (NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA, ''),
    (NetworkSysTypeEnum.OUTPUT_DATA_AGGREGATOR_DEVICE_DATA, ''),
}


class DeviceDataStruct(NamedTuple):
    """
    Data structure of payload, packets and sending recipes for workers

    Attributes
    ----------
    payload : PayloadInfo
        data structure of payload for workers
    sending_recipes : List[DeviceDataStructSendingRecipe]
        array of data structures of sending recipes for workers
    device_packets : List[DevicePacket]
        array of data structures of packets for workers

    Methods
    -------
    send(sending_map: Mapping[NetworkSysTypeEnum, Callable[[UniWorker, OutputMessage], None]], self_worker: Optional[UniWorker] = None) -> None:
        Preparation of sending recipes for workers
        Filtering recipes for OUTPUT_DATA_LOGGER_DEVICE_DATA which must be sent with empty 'specifier' and 'params'
    """
    payload: PayloadInfo
    sending_recipes: List[DeviceDataStructSendingRecipe]
    device_packets: List[DevicePacket]

    def send(
        self,
        sending_map: Mapping[NetworkSysTypeEnum, Callable[[Optional[UniWorker[UniMessage, None]], OutputMessage], None]],
        *,
        self_worker: Optional[UniWorker[UniMessage, None]] = None,
    ) -> None:
        """
        This method preparing sending recipes for workers and send them to worker which defined by 'sending_map' param.

        Parameters
        ----------
        sending_map : Mapping[NetworkSysTypeEnum, Callable[[Optional[UniWorker[UniMessage, None]], OutputMessage], None]]
            Sending map for identifying of worker name to whom must be sent the message
        self_worker : Optional[UniWorker[UniMessage, None]] = None
            Worker object (Optional)

        Returns
        -------
        None
        """
        assert set(sending_map.keys()) <= ALL_SUPPORTED_OUTPUT_NETWORK_TYPES
        for recipe in self.sending_recipes:
            if recipe.net_sys_type is not NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA and recipe.net_sys_type in sending_map:
                assert recipe.network is not None
                sending_map[recipe.net_sys_type](
                    self_worker,
                    OutputMessage(
                        data=recipe.message,
                        specifier=recipe.network.specifier or '',
                        network_id=recipe.network.id,
                        params=recipe.network.params or {},
                    ),
                )
            else:
                sending_map[NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA](
                    self_worker,
                    OutputMessage(
                        data=recipe.message,
                        specifier='',
                        params={},
                    ),
                )


class DeviceDataStructBatch(NamedTuple):
    """
    Data structure which contains array of other data structures of payloads, packets and sending recipes for workers

    Attributes
    ----------
    messages : List[DeviceDataStruct]
        array of data structures of payloads, packets and sending recipes for workers

    Methods
    -------
    send(self_worker: UniWorker, sending_map: Mapping[NetworkSysTypeEnum, Callable[[UniWorker, OutputMessage | List[OutputMessage]], None]]) -> None:
        Preparation of sending recipes for workers
        Filtering recipes for OUTPUT_DATA_AGGREGATOR_DEVICE_DATA and storing it to other array for sending batch of [OutputMessage]
    """
    messages: List[DeviceDataStruct]

    def send(
        self,
        sending_map: Mapping[NetworkSysTypeEnum, Callable[[Optional[UniWorker[UniMessage, None]], OutputMessage | List[OutputMessage]], None]],
        *,
        self_worker: Optional[UniWorker[UniMessage, None]] = None,
    ) -> None:
        """
        This method preparing sending recipes for workers and send them to worker which defined by 'sending_map' param.

        If network is OUTPUT_DATA_AGGREGATOR_DEVICE_DATA then put recipe to other array to send them all to batch worker
            in other case it uses usual filtering from DeviceDataStruct.send method

        Parameters
        ----------
        sending_map : Mapping[NetworkSysTypeEnum, Callable[[Optional[UniWorker[UniMessage, None]], OutputMessage | List[OutputMessage]], None]]
            Sending map for identifying of worker name to whom must be sent the message
        self_worker : Optional[UniWorker[UniMessage, None]] = None
            Worker object (Optional)

        Returns
        -------
        None
        """
        if set(sending_map.keys()) > ALL_SUPPORTED_OUTPUT_NETWORK_TYPES_BATCH:
            logger.error('Sending map keys are not equal to SUPPORTED NETWORK TYPES! :: skip')
            return
        da_output_batch: List[OutputMessage] = []
        redirect_network_output_batch: List[OutputMessage] = []
        for message in self.messages:
            for recipe in message.sending_recipes:
                if recipe.net_sys_type is NetworkSysTypeEnum.OUTPUT_DATA_AGGREGATOR_DEVICE_DATA:
                    da_output_batch.append(OutputMessage(
                        data=recipe.message,
                        specifier=recipe.network.specifier or '',  # type: ignore
                        params=recipe.network.params or {},  # type: ignore
                    ))
                elif recipe.net_sys_type is NetworkSysTypeEnum.OUTPUT_REDIRECT_NETWORK:
                    redirect_network_output_batch.append(OutputMessage(
                        data=recipe.message,
                        specifier=recipe.network.specifier or '',  # type: ignore
                        params=recipe.network.params or {},  # type: ignore
                    ))
                else:
                    if recipe.net_sys_type is not NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA and recipe.net_sys_type in sending_map:
                        assert recipe.network is not None
                        sending_map[recipe.net_sys_type](
                            self_worker,
                            OutputMessage(
                                data=recipe.message,
                                specifier=recipe.network.specifier or '',
                                params=recipe.network.params or {},
                            ),
                        )
                    else:
                        sending_map[NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA](
                            self_worker,
                            OutputMessage(
                                data=recipe.message,
                                specifier='',
                                params={},
                            ),
                        )

        if da_output_batch and NetworkSysTypeEnum.OUTPUT_DATA_AGGREGATOR_DEVICE_DATA in sending_map:
            logger.info(f'recipes output DA: {da_output_batch}')
            sending_map[NetworkSysTypeEnum.OUTPUT_DATA_AGGREGATOR_DEVICE_DATA](self_worker, da_output_batch)
        if redirect_network_output_batch and NetworkSysTypeEnum.OUTPUT_REDIRECT_NETWORK in sending_map:
            logger.info(f'recipes output REDIRECT_NETWORK: {redirect_network_output_batch}')
            grouper_params = lambda obj: json.dumps(obj.params)  # noqa: E731
            sorted_payload_by_params = sorted(redirect_network_output_batch, key=grouper_params)
            for key_params, sorted_messages in groupby(sorted_payload_by_params, key=grouper_params):
                logger.info(f'recipes output REDIRECT_NETWORK with params: {key_params} :: {sorted_messages}')
                sending_map[NetworkSysTypeEnum.OUTPUT_REDIRECT_NETWORK](self_worker, list(sorted_messages))


class DeviceDataManager(NamedTuple):
    raw_mac: int
    rewritten_mac: int
    kind: 'DeviceDataManagerKind'
    gateway_id: UUID
    current_network: DataGatewayNetwork
    send_always_to_data_logger: bool
    device_data: Optional[DADeviceData]
    current_net_rel: Optional[ApiDataGatewayNetworkDevice | ApiDataGatewayNetworkOptimized]
    default_networks: List[DataGatewayNetwork] = []
    error: IntegrationV0MessageError = IntegrationV0MessageError()

    def make_struct_batch(
        self,
        *,
        payload: List[IntegrationV0MessageGateway],
    ) -> DeviceDataStructBatch:
        device_pack: List[DevicePacket] = []
        messages_batch: List[DeviceDataStruct] = []
        decrypted_payload_hex: str
        logger.info(f'payload type: list of {type(payload[0]).__name__}' if payload else 'payload type: empty list')
        base_messages: List[IntegrationV0MessageGateway] = payload.copy()
        logger.info('payload was copied from input')

        for base_message in base_messages:
            decrypted_payload_hex = base_message.decrypted_payload
            encryption_type = None
            sending_recipes: List[DeviceDataStructSendingRecipe] = []
            payload_str = base_message.raw_payload
            if payload_str != '':
                encryption_type = None
                decrypted_payload = bytes([])
                payload_item = bytes.fromhex(payload_str)
                capture_dt = base_message.date_created

                if self.device_data is not None:
                    assert self.current_net_rel is not None  # only for mypy
                    if self.kind is DeviceDataManagerKind.UPLINK:
                        encryption_type = EncryptionType(self.current_net_rel.uplink_encryption_type.value)
                        key = base64.b64decode(self.current_net_rel.uplink_encryption_key) if self.current_net_rel.uplink_encryption_key else b''
                    else:
                        encryption_type = EncryptionType(self.current_net_rel.downlink_encryption_type.value)
                        key = base64.b64decode(self.current_net_rel.downlink_encryption_key) if self.current_net_rel.downlink_encryption_key else b''
                    try:
                        decrypted_payload = encryption_type.decrypt(payload_item, key)
                        logger.info('payload decrypted')
                    except Exception as e:  # noqa: B902
                        self.error.error_message = str(e)
                        self.error.error_type = IntegrationV0MessageErrorType.data_undecryptable
                        logger.warning(f'payload decryption error :: {e}')
                    else:
                        try:
                            device_pack = DeviceProtocolType(self.current_net_rel.protocol.type.value).parse(decrypted_payload)
                            logger.info('payload parsed')
                        except DataGatewayDeviceProtocolParsingError as e:
                            self.error.error_message = str(e)
                            self.error.error_type = IntegrationV0MessageErrorType.data_unparsable
                            logger.warning(f'payload parsing error {e}')

                decrypted_payload_hex = decrypted_payload.hex()
                server_time = datetime.utcnow().replace(tzinfo=timezone.utc)
                if capture_dt.tzinfo:
                    capture_dt = capture_dt.astimezone(timezone.utc).replace(tzinfo=timezone.utc)
                else:
                    capture_dt = capture_dt.replace(tzinfo=timezone.utc)
                if capture_dt > (server_time + timedelta(minutes=2)):
                    self.error.error_message = f'Message time is in future {capture_dt}, server time {server_time}'
                    self.error.error_type = IntegrationV0MessageErrorType.packet_from_the_future
                elif capture_dt < (server_time - timedelta(days=365)):
                    self.error.error_message = f'Message time is older than 1 year from now {capture_dt}, server time {server_time}'
                    self.error.error_type = IntegrationV0MessageErrorType.packet_from_the_past
                hacks: List[DeviceHack] = []
                device_tz = pytz.timezone('UTC')
                if self.device_data is not None:
                    hacks = getattr(self.device_data, 'hacks', [])
                    if self.device_data.device_tz is not None:
                        device_tz = pytz.timezone(self.device_data.device_tz)

                integration_message_datas: List[Tuple[IntegrationV0MessageData, str]] = [
                    (
                        integration_mes,
                        type(pack.packet).__name__,  # type: ignore
                    ) for pack in device_pack for integration_mes in pack.to_integration_data(
                        received_at=capture_dt,
                        device_tz=device_tz,
                        hacks=hacks,
                    )]
                if not integration_message_datas:
                    integration_message_datas = [(IntegrationV0MessageData(dt=capture_dt), 'EmptyDevicePacket')]

                parsed_base_messages = [IntegrationV0MessageGateway(
                    date_created=capture_dt,
                    data=integration_message_data,
                    meta=base_message.meta,
                    packet_type_name=packet_type_name,
                    id=base_message.id,
                    device_mac=self.rewritten_mac,
                    raw_payload=payload_str,
                    decrypted_payload=decrypted_payload_hex,
                    device_raw_mac=self.raw_mac,
                    protocol_name=self.current_net_rel.protocol.name if self.current_net_rel else None,
                    protocol_type=self.current_net_rel.protocol.type if self.current_net_rel else None,
                    raw_message=json.dumps(base_message, cls=CustomJSONEncoder),
                    gateway_id=self.gateway_id,
                    network_id=self.current_network.id,
                    network_sys_type=self.current_network.sys_type,
                    device_id=self.current_net_rel.device_id if self.current_net_rel else None,
                    protocol_id=self.current_net_rel.protocol.id if self.current_net_rel else None,
                    verified_device=self.device_data is not None,
                    dt_calculated=True if capture_dt != integration_message_data.dt else False,
                    error=self.error,
                ) for integration_message_data, packet_type_name in integration_message_datas]

                if self.send_always_to_data_logger:
                    for parsed_base_message in parsed_base_messages:
                        sending_recipes.append(DeviceDataStructSendingRecipe(
                            network=None,
                            message=parsed_base_message.model_copy(),
                            net_sys_type=NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA,
                        ))
                if self.device_data:
                    for net_rel in self.device_data.all_data_gateway_networks:  # TODO: data-aggregator setting for sending for repeater
                        if net_rel.network.type_network is not NetworkTypeEnum.output:
                            continue

                        if net_rel.network.sys_type not in ALL_SUPPORTED_OUTPUT_NETWORK_TYPES:
                            continue

                        if self.send_always_to_data_logger and net_rel.network.sys_type is NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA:
                            continue

                        for parsed_base_message in parsed_base_messages:
                            message = parsed_base_message.model_copy()
                            message.device_mac = net_rel.mac
                            message.device_id = net_rel.device_id

                            sending_recipes.append(DeviceDataStructSendingRecipe(
                                network=net_rel.network,
                                message=message,
                                net_sys_type=net_rel.network.sys_type,
                            ))
                else:
                    for net in self.default_networks:
                        for parsed_base_message in parsed_base_messages:
                            sending_recipes.append(DeviceDataStructSendingRecipe(
                                network=net,
                                message=parsed_base_message.model_copy(),
                                net_sys_type=net.sys_type,
                            ))

                messages_batch.append(DeviceDataStruct(
                    sending_recipes=sending_recipes,
                    payload=PayloadInfo(encrypted_payload=payload_str, decrypted_payload=decrypted_payload_hex or None, encryption_type=encryption_type),
                    device_packets=device_pack,
                ))
            else:
                if self.send_always_to_data_logger:
                    sending_recipes.append(DeviceDataStructSendingRecipe(
                        network=None,
                        message=base_message.model_copy(),
                        net_sys_type=NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA,
                    ))
                if self.device_data:
                    for net_rel in self.device_data.all_data_gateway_networks:  # TODO: data-aggregator setting for sending for repeater
                        if net_rel.network.type_network is not NetworkTypeEnum.output:
                            continue

                        if net_rel.network.sys_type not in ALL_SUPPORTED_OUTPUT_NETWORK_TYPES:
                            continue

                        if self.send_always_to_data_logger and net_rel.network.sys_type is NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA:
                            continue

                        message = base_message.model_copy()
                        message.device_mac = net_rel.mac
                        message.device_id = net_rel.device_id

                        sending_recipes.append(DeviceDataStructSendingRecipe(
                            network=net_rel.network,
                            message=message,
                            net_sys_type=net_rel.network.sys_type,
                        ))
                else:
                    for net in self.default_networks:
                        sending_recipes.append(DeviceDataStructSendingRecipe(
                            network=net,
                            message=base_message.model_copy(),
                            net_sys_type=net.sys_type,
                        ))

                messages_batch.append(DeviceDataStruct(
                    sending_recipes=sending_recipes,
                    payload=PayloadInfo(encrypted_payload=payload_str, decrypted_payload=decrypted_payload_hex or None, encryption_type=encryption_type),
                    device_packets=device_pack,
                ))

        return DeviceDataStructBatch(
            messages=messages_batch,
        )

    def make_struct(
        self,
        *,
        payload: Union[bytes, IntegrationV0MessageGateway],
        capture_dt: datetime,
        meta: IntegrationV0MessageMeta,
        raw: str,
        message_id: Optional[UUID],
        bs_geo: Optional[BaseStationInfoGeo] = None,
    ) -> DeviceDataStruct:
        device_pack: List[DevicePacket] = []
        base_messages = []
        logger.info(f'payload type of {type(payload)}')
        if isinstance(payload, IntegrationV0MessageGateway):
            base_message = payload.model_copy()
            base_message.meta = meta
            base_message.error = self.error
            base_messages.append(base_message)
            decrypted_payload_hex = payload.decrypted_payload
            encryption_type = None
            payload_str = payload.raw_payload
            logger.info('payload was copied from input')
        else:
            encryption_type = None
            decrypted_payload = bytes([])
            if self.device_data is not None:
                assert self.current_net_rel is not None  # only for mypy
                if self.kind is DeviceDataManagerKind.UPLINK:
                    encryption_type = EncryptionType(self.current_net_rel.uplink_encryption_type.value)
                    key = base64.b64decode(self.current_net_rel.uplink_encryption_key) if self.current_net_rel.uplink_encryption_key else b''
                else:
                    encryption_type = EncryptionType(self.current_net_rel.downlink_encryption_type.value)
                    key = base64.b64decode(self.current_net_rel.downlink_encryption_key) if self.current_net_rel.downlink_encryption_key else b''
                try:
                    decrypted_payload = encryption_type.decrypt(payload, key)
                    logger.info('payload decrypted')
                except Exception as e:  # noqa: B902
                    self.error.error_message = str(e)
                    self.error.error_type = IntegrationV0MessageErrorType.data_undecryptable
                    logger.warning(f'payload decryption error :: {e}')
                else:
                    try:
                        device_pack = DeviceProtocolType(self.current_net_rel.protocol.type.value).parse(decrypted_payload)
                        logger.info('payload parsed')
                    except DataGatewayDeviceProtocolParsingError as e:
                        self.error.error_message = str(e)
                        self.error.error_type = IntegrationV0MessageErrorType.data_unparsable
                        logger.warning(f'payload parsing error {e}')

            decrypted_payload_hex = decrypted_payload.hex()
            server_time = datetime.utcnow().replace(tzinfo=timezone.utc)
            if capture_dt.tzinfo:
                capture_dt = capture_dt.astimezone(timezone.utc).replace(tzinfo=timezone.utc)
            else:
                capture_dt = capture_dt.replace(tzinfo=timezone.utc)
            if capture_dt > (server_time + timedelta(minutes=2)):
                self.error.error_message = f'Message time is in future {capture_dt}, server time {server_time}'
                self.error.error_type = IntegrationV0MessageErrorType.packet_from_the_future
            elif capture_dt < (server_time - timedelta(days=365)):
                self.error.error_message = f'Message time is older than 1 year from now {capture_dt}, server time {server_time}'
                self.error.error_type = IntegrationV0MessageErrorType.packet_from_the_past
            hacks: List[DeviceHack] = []
            device_tz = pytz.timezone('UTC')
            if self.device_data is not None:
                hacks = getattr(self.device_data, 'hacks', [])
                if self.device_data.device_tz is not None:
                    device_tz = pytz.timezone(self.device_data.device_tz)
                elif bs_geo is not None:
                    device_tz = ZoneInfo(_timezone_finder.certain_timezone_at(lat=bs_geo.latitude, lng=bs_geo.longitude))  # type: ignore

            integration_message_datas: List[Tuple[IntegrationV0MessageData, str]] = [
                (
                    integration_mes,
                    type(pack.packet).__name__,  # type: ignore
                ) for pack in device_pack for integration_mes in pack.to_integration_data(
                    received_at=capture_dt,
                    device_tz=device_tz,
                    hacks=hacks,
                )]
            if not integration_message_datas:
                integration_message_datas = [(IntegrationV0MessageData(dt=capture_dt), 'EmptyDevicePacket')]

            for integration_message_data, packet_type_name in integration_message_datas:
                base_messages.append(IntegrationV0MessageGateway(
                    date_created=capture_dt,
                    data=integration_message_data,
                    meta=meta,
                    packet_type_name=packet_type_name,
                    id=message_id,
                    device_mac=self.rewritten_mac,
                    raw_payload=payload.hex(),
                    decrypted_payload=decrypted_payload_hex,
                    device_raw_mac=self.raw_mac,
                    protocol_name=self.current_net_rel.protocol.name if self.current_net_rel else None,
                    protocol_type=self.current_net_rel.protocol.type if self.current_net_rel else None,
                    raw_message=raw,
                    gateway_id=self.gateway_id,
                    network_id=self.current_network.id,
                    network_sys_type=self.current_network.sys_type,
                    device_id=self.current_net_rel.device_id if self.current_net_rel else None,
                    protocol_id=self.current_net_rel.protocol.id if self.current_net_rel else None,
                    verified_device=self.device_data is not None,
                    dt_calculated=True if capture_dt != integration_message_data.dt else False,
                    error=self.error,
                ))
            payload_str = payload.hex()

        sending_recipes: List[DeviceDataStructSendingRecipe] = []
        if self.send_always_to_data_logger:
            for base_message in base_messages:
                sending_recipes.append(DeviceDataStructSendingRecipe(
                    network=None,
                    message=base_message,
                    net_sys_type=NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA,
                ))
        if self.device_data:
            for net_rel in self.device_data.all_data_gateway_networks:  # TODO: data-aggregator setting for sending for repeater
                if net_rel.network.type_network is not NetworkTypeEnum.output:
                    continue

                if net_rel.network.sys_type not in ALL_SUPPORTED_OUTPUT_NETWORK_TYPES:
                    continue

                if self.send_always_to_data_logger and net_rel.network.sys_type is NetworkSysTypeEnum.OUTPUT_DATA_LOGGER_DEVICE_DATA:
                    continue

                for base_message in base_messages:
                    message = base_message.model_copy()
                    message.device_mac = net_rel.mac
                    message.device_id = net_rel.device_id

                    sending_recipes.append(DeviceDataStructSendingRecipe(
                        network=net_rel.network,
                        message=message,
                        net_sys_type=net_rel.network.sys_type,
                    ))
        else:
            for net in self.default_networks:
                for base_message in base_messages:
                    sending_recipes.append(DeviceDataStructSendingRecipe(
                        network=net,
                        message=base_message.model_copy(),
                        net_sys_type=net.sys_type,
                    ))

        return DeviceDataStruct(
            sending_recipes=sending_recipes,
            payload=PayloadInfo(encrypted_payload=payload_str, decrypted_payload=decrypted_payload_hex or None, encryption_type=encryption_type),
            device_packets=device_pack,
        )


@unique
class DeviceDataManagerKind(Enum):
    UPLINK = 'UPLINK'
    DOWNLINK = 'DOWNLINK'

    def _get_gateway_networks(
        self,
        gateway_id: UUID,
        *,
        data_aggregator_sdk: DataAggregatorApiSdk,
    ) -> List[ApiDataGatewayNetworkResponse]:
        cached_networks_data = network_redis_manager.get_gateway_networks_cache(gateway_id)
        
        if cached_networks_data is not None:
            try:
                networks_list = [
                    ApiDataGatewayNetworkResponse.model_validate(net_dict)
                    for net_dict in cached_networks_data
                ]
                logger.info(
                    f'Using cached gateway networks for gateway_id={gateway_id}, '
                )
                return networks_list
            except Exception as e:
                logger.warning(
                    f'Failed to deserialize cached networks for gateway_id={gateway_id}: {e}, '
                    f'falling back to API call'
                )
        
        networks_list = list(data_aggregator_sdk.get_data_gateway_networks(gateway_id, None, None, None, None))
        
        try:
            networks_dicts = [net.model_dump(mode='json') for net in networks_list]
            network_redis_manager.set_gateway_networks_cache(gateway_id, networks_dicts, ttl_seconds=180)
            logger.debug(
                f'Cached gateway networks for gateway_id={gateway_id}, '
                f'networks_count={len(networks_list)}'
            )
        except Exception as e:
            logger.warning(f'Failed to cache gateway networks for gateway_id={gateway_id}: {e}')
        
        return networks_list

    def prepare(
        self,
        gateway_id: UUID,
        network_id: UUID,
        macs: List[int],
        *,
        data_aggregator_sdk: DataAggregatorApiSdk,
        send_always_to_data_logger: bool,
        rewrite_mac: Optional[Callable[[int], Tuple[bool, int]]] = None,
        rewrite_enc: Optional[Callable[[int], Optional[Tuple[str, str]]]] = None,
    ) -> Dict[int, 'DeviceDataManager']:
        raw_to_new_macs_map: Dict[int, int] = {}
        rewritten_macs: Set[int] = set()
        forbidden_macs: Set[int] = set()
        raw_mac_to_enc_map: Dict[int, Tuple[EncryptionType, bytes]] = {}
        raw_macs: Set[int] = set()
        for mac in macs:
            forbidden, new_mac = rewrite_mac(mac) if rewrite_mac is not None else (False, mac)
            enc = None
            if rewrite_enc is not None:
                if enc := rewrite_enc(mac):
                    raw_mac_to_enc_map[new_mac] = EncryptionType(enc[0]), base64.b64decode(enc[1])
            if mac != new_mac:
                if not enc:
                    raw_macs.add(mac)
            raw_to_new_macs_map[mac] = new_mac
            rewritten_macs.add(new_mac)
            if forbidden:
                forbidden_macs.add(mac)
        default_networks: List[DataGatewayNetwork] = []
        current_network = None
        new_devices_data: List[DADeviceData] = []
        old_devices_data: List[DADeviceData] = []
        try:
            networks_list = self._get_gateway_networks(gateway_id, data_aggregator_sdk=data_aggregator_sdk)
            
            for net in networks_list:
                if net.type_network == NetworkTypeEnum.output and net.is_alive:
                    net_params = net.params if isinstance(net.params, dict) else {}
                    if net_params.get('is_default', False):
                        default_networks.append(DataGatewayNetwork(
                            id=net.id,
                            name=net.name,
                            type_network=net.type_network,
                            sys_type=net.sys_type,
                            specifier=net.specifier,
                            params=net.params,
                        ))
                if net.id == network_id:
                    current_network = DataGatewayNetwork(
                        id=net.id,
                        name=net.name,
                        type_network=net.type_network,
                        sys_type=net.sys_type,
                        specifier=net.specifier,
                        params=net.params,
                    )

            new_devices_data = data_aggregator_sdk.get_devices_data_optimized(gateway_id, network_id, list(rewritten_macs)) if len(rewritten_macs) > 0 else []
            old_devices_data = data_aggregator_sdk.get_devices_data_optimized(gateway_id, network_id, list(raw_macs)) if len(raw_macs) > 0 else []

        except DataAggregatorRequestError as e:
            logger.info(f'data_aggregator_sdk.get_devices_data_optimized -> invalid: {e.args}')
        except DataAggregatorResponseError as e:
            logger.error(f'data_aggregator_sdk.get_devices_data_optimized -> failed: {e}')
            raise

        manager_by_old_mac: Dict[int, DeviceDataManager] = {}
        for mac in macs:
            error = IntegrationV0MessageError()
            if mac in forbidden_macs:
                logger.error(msg := f'Mac {mac} is forbidden for current bs serial number, packet is skipped')
                error = IntegrationV0MessageError(
                    error_message=msg,
                    error_type=IntegrationV0MessageErrorType.mac_duplicated,
                )

            device_data = _search_device_data(new_devices_data, gateway_id, network_id, raw_to_new_macs_map[mac])
            if device_data is None:
                dev_data = None
                cur_net_rel = None
                logger.info(msg := f'mac: {raw_to_new_macs_map.get(mac, mac)} data not found in: {network_id}')
                if error.error_type == IntegrationV0MessageErrorType.none:
                    error = IntegrationV0MessageError(
                        error_message=msg,
                        error_type=IntegrationV0MessageErrorType.device_unidentified,
                    )
            else:
                logger.info(f'mac: {mac} FOUND in data aggregator in network: {network_id}')
                dev_data, cur_net_rel = device_data

                enc_type: Optional[DaEncryptionType] = None
                enc_key_bytes: Optional[str] = None
                if mac in raw_mac_to_enc_map:
                    enc_type, _enc_key_bytes = raw_mac_to_enc_map[mac]  # type: ignore
                    enc_key_bytes = _enc_key_bytes.decode('utf-8')
                elif mac in raw_macs:
                    _old_device_data = _search_device_data(old_devices_data, gateway_id, network_id, mac)
                    if _old_device_data is not None:
                        _1, old_curr_net_rel = _old_device_data
                        if self is DeviceDataManagerKind.UPLINK:
                            enc_type, enc_key_bytes = old_curr_net_rel.uplink_encryption_type, old_curr_net_rel.uplink_encryption_key
                        else:
                            enc_type, enc_key_bytes = old_curr_net_rel.downlink_encryption_type, old_curr_net_rel.downlink_encryption_key
                if enc_type is not None:
                    if self is DeviceDataManagerKind.UPLINK:
                        cur_net_rel.uplink_encryption_type = enc_type
                        cur_net_rel.uplink_encryption_key = enc_key_bytes
                    # else:  # ToDo: downlink encryption
                    #     cur_net_rel.downlink_encryption_type = enc_type
                    #     cur_net_rel.downlink_encryption_key = enc_key_bytes
            assert current_network is not None      # just for mypy
            manager_by_old_mac[mac] = DeviceDataManager(
                raw_mac=mac,
                rewritten_mac=raw_to_new_macs_map[mac],
                kind=self,
                gateway_id=gateway_id,
                current_network=current_network,
                device_data=dev_data,
                current_net_rel=cur_net_rel,
                default_networks=default_networks,
                send_always_to_data_logger=send_always_to_data_logger,
                error=error,
            )
        return manager_by_old_mac


assert len(DeviceDataManagerKind) == 2, 'should have only 2 values. because logic support only 2'


def _search_device_data(devices_data: List[DADeviceData], gateway_id: UUID, network_id: UUID, mac: int) -> Optional[Tuple[DADeviceData, ApiDataGatewayNetworkOptimized]]:
    for device_data in devices_data:
        for net_rel in device_data.all_data_gateway_networks:
            if net_rel.mac == mac and network_id == net_rel.network.id and net_rel.network.data_gateway.id == gateway_id:
                return device_data, net_rel
    return None



from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_08b_daily_new import \
    SmpmUlDeviceWaterMeter08BDailyNewData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_water_meter_08b_daily_new() -> None:
    case_serialized = bytes.fromhex("74e4eab1d6858f87")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=112323.3, temperature=23, battery_volts=3.3, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=False, value_reverse=1.084) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(value=112323.3, temperature=23, battery_volts=3.3, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=False, value_reverse=1111.1))  # noqa: E501
    case_serialized = bytes.fromhex("74000000f8031f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=92, battery_volts=0.0, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=0.0, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=92, value=0.0, value_reverse=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("74000000f8031f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=92, battery_volts=0.0, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=0.0, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=92, value=0.0, value_reverse=4.096))  # noqa: E501
    case_serialized = bytes.fromhex("74000000f8031f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=92, battery_volts=0.0, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=0.0, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=92, value=268435.456, value_reverse=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("74000000f8031f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=92, battery_volts=0.0, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=0.0, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=92, value=268435.456, value_reverse=4.096))  # noqa: E501
    case_serialized = bytes.fromhex("7400000000001f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=-35, battery_volts=0.0, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=0.0, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=-35, value=0.0, value_reverse=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("7400000000001f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=-35, battery_volts=0.0, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=0.0, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=-35, value=0.0, value_reverse=4.096))  # noqa: E501
    case_serialized = bytes.fromhex("7400000000001f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=-35, battery_volts=0.0, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=0.0, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=-35, value=268435.456, value_reverse=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("7400000000001f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=-35, battery_volts=0.0, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=0.0, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=-35, value=268435.456, value_reverse=4.096))  # noqa: E501
    case_serialized = bytes.fromhex("74000000f8ff1f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=92, battery_volts=6.3, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=6.3, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=92, value=0.0, value_reverse=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("74000000f8ff1f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=92, battery_volts=6.3, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=6.3, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=92, value=0.0, value_reverse=4.096))  # noqa: E501
    case_serialized = bytes.fromhex("74000000f8ff1f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=92, battery_volts=6.3, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=6.3, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=92, value=268435.456, value_reverse=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("74000000f8ff1f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=92, battery_volts=6.3, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=6.3, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=92, value=268435.456, value_reverse=4.096))  # noqa: E501
    case_serialized = bytes.fromhex("7400000000fc1f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=-35, battery_volts=6.3, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=6.3, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=-35, value=0.0, value_reverse=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("7400000000fc1f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=-35, battery_volts=6.3, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=6.3, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=-35, value=0.0, value_reverse=4.096))  # noqa: E501
    case_serialized = bytes.fromhex("7400000000fc1f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=-35, battery_volts=6.3, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=6.3, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=-35, value=268435.456, value_reverse=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("7400000000fc1f00")  # overflow
    assert SmpmUlDeviceWaterMeter08BDailyNewData(value=0.0, temperature=-35, battery_volts=6.3, event_low_battery_level=True, event_temperature_limits=True, event_reset=True, event_magnet=True, event_reverse_flow=True, value_reverse=0.0) == SmpmUlDeviceWaterMeter08BDailyNewData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BDailyNewData.serialize(SmpmUlDeviceWaterMeter08BDailyNewData(battery_volts=6.3, event_low_battery_level=True, event_magnet=True, event_reset=True, event_reverse_flow=True, event_temperature_limits=True, temperature=-35, value=268435.456, value_reverse=4.096))  # noqa: E501



from datetime import datetime, timedelta

import pytest

from data_gateway_sdk.utils.timestamp_calculation import timestamp_calculation


@pytest.mark.parametrize(
    'current_pack_time,pack_timestamp,overflow_clock_value_timedelta,expected',
    (
        (
            datetime(year=2023, month=3, day=7),
            timedelta(seconds=int('01111110101011101100111000', 2)),
            timedelta(seconds=int('11111111111111111111111111', 2)),
            datetime(2023, 3, 7, 2, 6, 47),
        ),
        (
            datetime(year=2023, month=1, day=11),
            timedelta(seconds=int('01101100101011101000111000', 2)),
            timedelta(seconds=int('11111111111111111111111111', 2)),
            datetime(2023, 1, 11, 11, 19, 19),
        ),
        (
            datetime(year=2022, month=1, day=2, hour=3),
            timedelta(seconds=int('11110001011111111000111000', 2)),
            timedelta(seconds=int('11111111111111111111111111', 2)),
            datetime(2022, 1, 2, 17, 22),
        ),
    ),
)
def test_timestamp_calculation_(current_pack_time: datetime, pack_timestamp: timedelta, overflow_clock_value_timedelta: timedelta, expected: datetime) -> None:
    assert timestamp_calculation(
        current_pack_time=current_pack_time,
        timestamp=pack_timestamp,
        overflow_clock_value_timedelta=overflow_clock_value_timedelta,
    ) == expected



from data_gateway_sdk.protocols.smpm.smpm_dl_device_energy_16b_get_data import SmpmDlDeviceEnergy16BGetDataData, SmpmDlDeviceEnergy16bGetDataDataRequestMonth, \
    SmpmDlDeviceEnergy16bGetDataDataRequestId
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_dl_device_energy_16b_get_data() -> None:
    case_serialized = bytes.fromhex("8101c5cb0c3e83c9a032e00d00000000")
    assert SmpmDlDeviceEnergy16BGetDataData(year=2032, month=SmpmDlDeviceEnergy16bGetDataDataRequestMonth.JAN, day=15, request_data_pack_ids=(SmpmDlDeviceEnergy16bGetDataDataRequestId.MONTHLY_ENERGY_ACTIVE_CONSUMED, SmpmDlDeviceEnergy16bGetDataDataRequestId.MONTHLY_ENERGY_REACTIVE_CONSUMED, SmpmDlDeviceEnergy16bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_3, SmpmDlDeviceEnergy16bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_SUM, SmpmDlDeviceEnergy16bGetDataDataRequestId.NETWORK_PARAMS_PHASE1, SmpmDlDeviceEnergy16bGetDataDataRequestId.UNDEFINED)) == SmpmDlDeviceEnergy16BGetDataData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("810130c80c3e83c9a032e00d00000000")  # overflow
    assert SmpmDlDeviceEnergy16BGetDataData(year=2000, month=SmpmDlDeviceEnergy16bGetDataDataRequestMonth.DEC, day=0, request_data_pack_ids=(SmpmDlDeviceEnergy16bGetDataDataRequestId.MONTHLY_ENERGY_ACTIVE_CONSUMED, SmpmDlDeviceEnergy16bGetDataDataRequestId.MONTHLY_ENERGY_REACTIVE_CONSUMED, SmpmDlDeviceEnergy16bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_3, SmpmDlDeviceEnergy16bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_SUM, SmpmDlDeviceEnergy16bGetDataDataRequestId.NETWORK_PARAMS_PHASE1, SmpmDlDeviceEnergy16bGetDataDataRequestId.UNDEFINED)) == SmpmDlDeviceEnergy16BGetDataData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("81f933c80c3e83c9a032e00d00000000")  # overflow
    assert SmpmDlDeviceEnergy16BGetDataData(year=2127, month=SmpmDlDeviceEnergy16bGetDataDataRequestMonth.DEC, day=0, request_data_pack_ids=(SmpmDlDeviceEnergy16bGetDataDataRequestId.MONTHLY_ENERGY_ACTIVE_CONSUMED, SmpmDlDeviceEnergy16bGetDataDataRequestId.MONTHLY_ENERGY_REACTIVE_CONSUMED, SmpmDlDeviceEnergy16bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_3, SmpmDlDeviceEnergy16bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_SUM, SmpmDlDeviceEnergy16bGetDataDataRequestId.NETWORK_PARAMS_PHASE1, SmpmDlDeviceEnergy16bGetDataDataRequestId.UNDEFINED)) == SmpmDlDeviceEnergy16BGetDataData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("8101f0cf0c3e83c9a032e00d00000000")  # overflow
    assert SmpmDlDeviceEnergy16BGetDataData(year=2000, month=SmpmDlDeviceEnergy16bGetDataDataRequestMonth.DEC, day=31, request_data_pack_ids=(SmpmDlDeviceEnergy16bGetDataDataRequestId.MONTHLY_ENERGY_ACTIVE_CONSUMED, SmpmDlDeviceEnergy16bGetDataDataRequestId.MONTHLY_ENERGY_REACTIVE_CONSUMED, SmpmDlDeviceEnergy16bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_3, SmpmDlDeviceEnergy16bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_SUM, SmpmDlDeviceEnergy16bGetDataDataRequestId.NETWORK_PARAMS_PHASE1, SmpmDlDeviceEnergy16bGetDataDataRequestId.UNDEFINED)) == SmpmDlDeviceEnergy16BGetDataData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("81f9f3cf0c3e83c9a032e00d00000000")  # overflow
    assert SmpmDlDeviceEnergy16BGetDataData(year=2127, month=SmpmDlDeviceEnergy16bGetDataDataRequestMonth.DEC, day=31, request_data_pack_ids=(SmpmDlDeviceEnergy16bGetDataDataRequestId.MONTHLY_ENERGY_ACTIVE_CONSUMED, SmpmDlDeviceEnergy16bGetDataDataRequestId.MONTHLY_ENERGY_REACTIVE_CONSUMED, SmpmDlDeviceEnergy16bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_3, SmpmDlDeviceEnergy16bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_SUM, SmpmDlDeviceEnergy16bGetDataDataRequestId.NETWORK_PARAMS_PHASE1, SmpmDlDeviceEnergy16bGetDataDataRequestId.UNDEFINED)) == SmpmDlDeviceEnergy16BGetDataData.parse(BufRef(case_serialized))  # noqa: E501



from datetime import datetime, tzinfo
from decimal import Decimal
from typing import Any, List, Dict

from data_aggregator_sdk.constants.enums import SensorType, IntegrationV0MessageEvent
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageSensor, \
    IntegrationV0MessageConsumption, CounterType, ResourceType, IntegrationV0MessageGeneration

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   smpm_ul_device_water_meter_08b_daily_new
#
# RESULT int:        9768173273944351860
# RESULT bin:  MSB   10000111 10001111 10000101 11010110 10110001 11101010 11100100 01110100   LSB
# RESULT hex:  LE    74 e4 ea b1 d6 85 8f 87
#
# name                      type    size  value(int)                                                        data(bits)
# ------------------------  ------  ----  ----------  ----------------------------------------------------------------
# packet_type_id.0.VAL      u7         7         116                                                           1110100
# packet_type_id.0.DFF      bool       1           0                                                          0
# value                     uf27p3    27   112323300                               110101100011110101011100100
# temperature               u7         7          58                        0111010
# battery_volts             uf6p1      6          33                  100001
# event_low_battery_level   bool       1           1                 1
# event_temperature_limits  bool       1           1                1
# event_reset               bool       1           1               1
# event_magnet              bool       1           1              1
# event_reverse_flow        bool       1           0             0
# value_reverse             uf11p3    11        1084  10000111100


class SmpmUlDeviceWaterMeter08BDailyNewData(Packet):
    value: float
    temperature: int
    battery_volts: float
    event_low_battery_level: bool
    event_temperature_limits: bool
    event_reset: bool
    event_magnet: bool
    event_reverse_flow: bool
    value_reverse: float

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((116) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.value, (int, float))
        result |= ((int(round(float(data.value) * 1000.0, 0)) & 134217727) & (2 ** (27) - 1)) << size
        size += 27
        assert isinstance(data.temperature, int)
        assert -35 <= data.temperature <= 92
        result |= (((data.temperature + 35)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 6.3
        result |= ((int(round(float(data.battery_volts) * 10.0, 0))) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.event_low_battery_level, bool)
        result |= ((int(data.event_low_battery_level)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_temperature_limits, bool)
        result |= ((int(data.event_temperature_limits)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_reset, bool)
        result |= ((int(data.event_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_magnet, bool)
        result |= ((int(data.event_magnet)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_reverse_flow, bool)
        result |= ((int(data.event_reverse_flow)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.value_reverse, (int, float))
        result |= ((int(round(float(data.value_reverse) * 1000.0, 0)) & 2047) & (2 ** (11) - 1)) << size
        size += 11
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceWaterMeter08BDailyNewData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 116 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["value"] = round(buf.shift(27) / 1000.0, 3)
        result__el_tmp1["temperature"] = buf.shift(7) + -35
        result__el_tmp1["battery_volts"] = round(buf.shift(6) / 10.0, 1)
        result__el_tmp1["event_low_battery_level"] = bool(buf.shift(1))
        result__el_tmp1["event_temperature_limits"] = bool(buf.shift(1))
        result__el_tmp1["event_reset"] = bool(buf.shift(1))
        result__el_tmp1["event_magnet"] = bool(buf.shift(1))
        result__el_tmp1["event_reverse_flow"] = bool(buf.shift(1))
        result__el_tmp1["value_reverse"] = round(buf.shift(11) / 1000.0, 3)
        result = SmpmUlDeviceWaterMeter08BDailyNewData(**result__el_tmp1)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(self.value)),
                        overloading_value=Decimal(str(134217.727)),
                    ),
                ],
                generation=[
                    IntegrationV0MessageGeneration(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        value=Decimal(str(self.value_reverse)),
                        overloading_value=Decimal(str(20.47)),
                    ),
                ],
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.BATTERY_IS_LOW] if self.event_low_battery_level else []),
                    *([IntegrationV0MessageEvent.TEMPERATURE_LIMIT] if self.event_temperature_limits else []),
                    *([IntegrationV0MessageEvent.RESET] if self.event_reset else []),
                    *([IntegrationV0MessageEvent.MAGNET_WAS_DETECTED] if self.event_magnet else []),
                    *([IntegrationV0MessageEvent.FLOW_REVERSE] if self.event_reverse_flow else []),
                ],
            ),
        ]



from data_gateway_sdk.protocols.smpm.smpm_ul_device_gaz_flow_08b_daily import SmpmUlDeviceGazFlow08BDailyData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_gaz_flow_08b_daily() -> None:
    case_serialized = bytes.fromhex("64e4eab1067b2100")
    assert SmpmUlDeviceGazFlow08BDailyData(cumulative_volume=112323.3, temperature=23, battery_volts=3.3, event_reset=False, event_case_was_opened=False, event_flow_reverse=False, event_flow_speed_is_over_limit=False, event_sensor_error_measurement=False, event_sensor_error_temperature=False, event_low_battery_level=False, event_system_error=False) == SmpmUlDeviceGazFlow08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceGazFlow08BDailyData.serialize(SmpmUlDeviceGazFlow08BDailyData(cumulative_volume=112323.3, temperature=23, battery_volts=3.3, event_reset=False, event_case_was_opened=False, event_flow_reverse=False, event_flow_speed_is_over_limit=False, event_sensor_error_measurement=False, event_sensor_error_temperature=False, event_low_battery_level=False, event_system_error=False))  # noqa: E501
    case_serialized = bytes.fromhex("6400000000ff807f")  # overflow
    assert SmpmUlDeviceGazFlow08BDailyData(cumulative_volume=0.0, temperature=155, battery_volts=0.0, event_reset=True, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_low_battery_level=True, event_system_error=True) == SmpmUlDeviceGazFlow08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceGazFlow08BDailyData.serialize(SmpmUlDeviceGazFlow08BDailyData(battery_volts=0.0, cumulative_volume=0.0, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_low_battery_level=True, event_reset=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_system_error=True, temperature=155))  # noqa: E501
    case_serialized = bytes.fromhex("640000000000807f")  # overflow
    assert SmpmUlDeviceGazFlow08BDailyData(cumulative_volume=0.0, temperature=-100, battery_volts=0.0, event_reset=True, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_low_battery_level=True, event_system_error=True) == SmpmUlDeviceGazFlow08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceGazFlow08BDailyData.serialize(SmpmUlDeviceGazFlow08BDailyData(battery_volts=0.0, cumulative_volume=0.0, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_low_battery_level=True, event_reset=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_system_error=True, temperature=-100))  # noqa: E501
    case_serialized = bytes.fromhex("6400000000ff807f")  # overflow
    assert SmpmUlDeviceGazFlow08BDailyData(cumulative_volume=0.0, temperature=155, battery_volts=0.0, event_reset=True, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_low_battery_level=True, event_system_error=True) == SmpmUlDeviceGazFlow08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceGazFlow08BDailyData.serialize(SmpmUlDeviceGazFlow08BDailyData(battery_volts=0.0, cumulative_volume=8589934.592, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_low_battery_level=True, event_reset=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_system_error=True, temperature=155))  # noqa: E501
    case_serialized = bytes.fromhex("640000000000807f")  # overflow
    assert SmpmUlDeviceGazFlow08BDailyData(cumulative_volume=0.0, temperature=-100, battery_volts=0.0, event_reset=True, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_low_battery_level=True, event_system_error=True) == SmpmUlDeviceGazFlow08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceGazFlow08BDailyData.serialize(SmpmUlDeviceGazFlow08BDailyData(battery_volts=0.0, cumulative_volume=8589934.592, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_low_battery_level=True, event_reset=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_system_error=True, temperature=-100))  # noqa: E501
    case_serialized = bytes.fromhex("6400000000ffff7f")  # overflow
    assert SmpmUlDeviceGazFlow08BDailyData(cumulative_volume=0.0, temperature=155, battery_volts=12.7, event_reset=True, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_low_battery_level=True, event_system_error=True) == SmpmUlDeviceGazFlow08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceGazFlow08BDailyData.serialize(SmpmUlDeviceGazFlow08BDailyData(battery_volts=12.7, cumulative_volume=0.0, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_low_battery_level=True, event_reset=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_system_error=True, temperature=155))  # noqa: E501
    case_serialized = bytes.fromhex("640000000000ff7f")  # overflow
    assert SmpmUlDeviceGazFlow08BDailyData(cumulative_volume=0.0, temperature=-100, battery_volts=12.7, event_reset=True, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_low_battery_level=True, event_system_error=True) == SmpmUlDeviceGazFlow08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceGazFlow08BDailyData.serialize(SmpmUlDeviceGazFlow08BDailyData(battery_volts=12.7, cumulative_volume=0.0, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_low_battery_level=True, event_reset=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_system_error=True, temperature=-100))  # noqa: E501
    case_serialized = bytes.fromhex("6400000000ffff7f")  # overflow
    assert SmpmUlDeviceGazFlow08BDailyData(cumulative_volume=0.0, temperature=155, battery_volts=12.7, event_reset=True, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_low_battery_level=True, event_system_error=True) == SmpmUlDeviceGazFlow08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceGazFlow08BDailyData.serialize(SmpmUlDeviceGazFlow08BDailyData(battery_volts=12.7, cumulative_volume=8589934.592, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_low_battery_level=True, event_reset=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_system_error=True, temperature=155))  # noqa: E501
    case_serialized = bytes.fromhex("640000000000ff7f")  # overflow
    assert SmpmUlDeviceGazFlow08BDailyData(cumulative_volume=0.0, temperature=-100, battery_volts=12.7, event_reset=True, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_low_battery_level=True, event_system_error=True) == SmpmUlDeviceGazFlow08BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceGazFlow08BDailyData.serialize(SmpmUlDeviceGazFlow08BDailyData(battery_volts=12.7, cumulative_volume=8589934.592, event_case_was_opened=True, event_flow_reverse=True, event_flow_speed_is_over_limit=True, event_low_battery_level=True, event_reset=True, event_sensor_error_measurement=True, event_sensor_error_temperature=True, event_system_error=True, temperature=-100))  # noqa: E501



from datetime import datetime
from uuid import UUID

from data_aggregator_sdk.integration_message import IntegrationV0MessageMeta
from pydantic import ValidationError

from data_gateway_sdk.errors import DataGatewayBSProtocolParsingError
from data_gateway_sdk.protocols.nero_bs_packet.http_nero_bs_packet import HttpV0NeroBsPacket, DevicePackage, \
    BaseStationInfo, BaseStationAdditionalInfo, BaseStationLogs, DevicePackageProtocolTypeEnum, \
    DevicePackageNetworkTypeEnum, BaseStationInfoGeo, BaseStationLogsTypeEnum, DataNbfi, DataUnbp, SdrInfo, \
    BaseStationEnvironmentInfo, DModEnum, DataUnbp2

DATETIME = datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11)

DEVICE_PACKET_NBFI = {
    'data': [
        {
            'dt': DATETIME,
            'raw': 'FFFFFFFFFFFFFFFF',
            'type': 'uplink',
            'protocol': 'nbfi',
            'data_nbfi': {'mac': 12345678, 'f_ask': True, 'iterator': 3, 'multi': False, 'system': True, 'payload': 'FFFFFFFFFFFFFFFF'},
            'sdr_info': {'freq': 870282100, 'freq_channel': 871282100, 'sdr': 0, 'baud_rate': 50, 'rssi': -75, 'snr': 65},
        },
    ],
    'base_station_environment_info': [
        {
            'dt': DATETIME,
            'spectrum': [0, 0, 0, 0, 0, 0],
            'sdr': 25,
            'freq_carrier': 0.1,
            'freq_delta': 0.5,
        },
    ],
    'base_station_info': {'geo': {'dt': DATETIME, 'actual': True}, 'uptime_s': 21},
    'base_station_additional_info': {'id': 15015},
    'base_station_logs': [{'type': 'info', 'dt': DATETIME, 'text': ''}],

}

DEVICE_PACKET_UNBP = {
    'data': [
        {
            'dt': DATETIME,
            'raw': 'FFFFFFFFFFFFFFFF',
            'type': 'downlink',
            'protocol': 'unbp',
            'data_unbp': {'mac': 12345678, 'ack': True, 'iterator': 3, 'payload': 'FFFFFFFF'},
            'sdr_info': {'freq': 870282100, 'freq_channel': 871282100, 'sdr': 0, 'baud_rate': 50, 'rssi': -75, 'snr': 65},
        },
    ],
    'base_station_environment_info': [
        {
            'dt': DATETIME,
            'spectrum': [0, 0, 0, 0, 0, 0],
            'sdr': 25,
            'freq_carrier': 0.1,
            'freq_delta': 0.5,
        },
    ],
    'base_station_info': {'geo': {'dt': DATETIME, 'actual': True}, 'uptime_s': 21},
    'base_station_additional_info': {'id': 15015},
    'base_station_logs': [{'type': 'info', 'dt': DATETIME, 'text': ''}],

}

DEVICE_PACKET_UNBP2 = {
    'data': [
        {
            'dt': DATETIME,
            'raw': 'FFFFFFFFFFFFFFFF',
            'type': 'downlink',
            'protocol': 'unbp2',
            'data_unbp2': {'mac': 12345678, 'payload': 'FFFFFFFF'},
            'sdr_info': {'freq': 870282100, 'freq_channel': 871282100, 'sdr': 0, 'baud_rate': 50, 'rssi': -75, 'snr': 65},
        },
    ],
    'base_station_environment_info': [
        {
            'dt': DATETIME,
            'spectrum': [0, 0, 0, 0, 0, 0],
            'sdr': 25,
            'freq_carrier': 0.1,
            'freq_delta': 0.5,
        },
    ],
    'base_station_info': {'geo': {'dt': DATETIME, 'actual': True}, 'uptime_s': 21},
    'base_station_additional_info': {'id': 15015},
    'base_station_logs': [{'type': 'info', 'dt': DATETIME, 'text': ''}],

}

DEVICE_PACKET_WRONG = {
    'data': [
        {
            'dt': datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
            'raw': 'FF',
            'type': 'downlink',
            'protocol': 'unbpp',
            'data_unbp': {'mac': 64, 'ack': True, 'iterator': 3, 'payload': 'FF'},
            'sdr_info': {'freq': 0, 'freq_channel': -1, 'sdr': 0, 'baud_rate': -1, 'rssi': -75, 'snr': 65},
        },
    ],
    'base_station_environment_info': [
        {
            'dt': datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
            'spectrum': [0, 0, 0, 0, 0, 0],
            'sdr': 25,
            'freq_carrier': 0.1,
            'freq_delta': 0.5,
        },
    ],
    'base_station_info': {
        'geo': {
            'dt': datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
            'actual': True,
        },
        'uptime_s': -1,
    },
    'base_station_additional_info': {'id': 15015},
    'base_station_logs': [
        {'type': 'info', 'dt': datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11), 'text': ''},
    ],

}

DEVICE_PACKAGE_NBFI = DevicePackage(
    dt=DATETIME,
    raw='FFFFFFFFFFFFFFFF',
    type=DevicePackageNetworkTypeEnum.uplink,
    protocol=DevicePackageProtocolTypeEnum.nbfi,
    data_nbfi=DataNbfi(mac=12345678, f_ask=True, iterator=3, multi=False, system=True, payload='FFFFFFFFFFFFFFFF'),
    sdr_info=SdrInfo(freq=870282100, freq_channel=871282100, sdr=0, baud_rate=50, rssi=-75, snr=65),
)

DEVICE_PACKAGE_UNBP = DevicePackage(
    dt=DATETIME,
    raw='FFFFFFFFFFFFFFFF',
    type=DevicePackageNetworkTypeEnum.downlink,
    protocol=DevicePackageProtocolTypeEnum.unbp,
    data_unbp=DataUnbp(mac=12345678, iterator=3, payload='FFFFFFFF', ack=True),
    sdr_info=SdrInfo(freq=870282100, freq_channel=871282100, sdr=0, baud_rate=50, rssi=-75, snr=65),
)

DEVICE_PACKAGE_UNBP2 = DevicePackage(
    dt=DATETIME,
    raw='FFFFFFFFFFFFFFFF',
    type=DevicePackageNetworkTypeEnum.downlink,
    protocol=DevicePackageProtocolTypeEnum.unbp2,
    data_unbp2=DataUnbp2(mac=12345678, payload='FFFFFFFF'),
    sdr_info=SdrInfo(freq=870282100, freq_channel=871282100, sdr=0, baud_rate=50, rssi=-75, snr=65),
)


def test_parsing_of_true_device_packet() -> None:
    assert DModEnum.DBPSK.__repr__() == 'DModEnum.DBPSK'
    assert DevicePackageNetworkTypeEnum.downlink.__repr__() == 'DevicePackageNetworkTypeEnum.downlink'
    assert BaseStationLogsTypeEnum.info.__repr__() == 'BaseStationLogsTypeEnum.info'

    assert type(DEVICE_PACKAGE_NBFI.payload_bytes) is bytes
    assert type(DEVICE_PACKAGE_UNBP.payload_bytes) is bytes
    assert type(DEVICE_PACKAGE_UNBP2.payload_bytes) is bytes
    assert type(DEVICE_PACKAGE_NBFI.mac) is int
    assert type(DEVICE_PACKAGE_UNBP.mac) is int
    assert type(DEVICE_PACKAGE_UNBP2.mac) is int

    assert HttpV0NeroBsPacket.parse(DEVICE_PACKET_NBFI) == HttpV0NeroBsPacket(
        data=[
            DEVICE_PACKAGE_NBFI,
        ],
        base_station_environment_info=[BaseStationEnvironmentInfo(
            dt=DATETIME,
            spectrum=[0, 0, 0, 0, 0, 0],
            sdr=25,
            freq_carrier=0.1,
            freq_delta=0.5,
        )],
        base_station_info=BaseStationInfo(geo=BaseStationInfoGeo(dt=DATETIME, actual=True), uptime_s=21),
        base_station_additional_info=BaseStationAdditionalInfo(id=15015),
        base_station_logs=[BaseStationLogs(type=BaseStationLogsTypeEnum.info, dt=DATETIME, text='')],
    )

    assert HttpV0NeroBsPacket.parse(DEVICE_PACKET_UNBP) == HttpV0NeroBsPacket(
        data=[
            DEVICE_PACKAGE_UNBP,
        ],
        base_station_environment_info=[BaseStationEnvironmentInfo(
            dt=datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
            spectrum=[0, 0, 0, 0, 0, 0],
            sdr=25,
            freq_carrier=0.1,
            freq_delta=0.5,
        )],
        base_station_info=BaseStationInfo(geo=BaseStationInfoGeo(dt=DATETIME, actual=True), uptime_s=21),
        base_station_additional_info=BaseStationAdditionalInfo(id=15015),
        base_station_logs=[BaseStationLogs(type=BaseStationLogsTypeEnum.info, dt=DATETIME, text='')],
    )

    assert HttpV0NeroBsPacket.parse(DEVICE_PACKET_UNBP2) == HttpV0NeroBsPacket(
        data=[
            DEVICE_PACKAGE_UNBP2,
        ],
        base_station_environment_info=[BaseStationEnvironmentInfo(
            dt=datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
            spectrum=[0, 0, 0, 0, 0, 0],
            sdr=25,
            freq_carrier=0.1,
            freq_delta=0.5,
        )],
        base_station_info=BaseStationInfo(geo=BaseStationInfoGeo(dt=DATETIME, actual=True), uptime_s=21),
        base_station_additional_info=BaseStationAdditionalInfo(id=15015),
        base_station_logs=[BaseStationLogs(type=BaseStationLogsTypeEnum.info, dt=DATETIME, text='')],
    )

    assert type(HttpV0NeroBsPacket(
        data=[
            DEVICE_PACKAGE_UNBP,
        ],
        base_station_environment_info=[BaseStationEnvironmentInfo(
            dt=datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
            spectrum=[0, 0, 0, 0, 0, 0],
            sdr=25,
            freq_carrier=0.1,
            freq_delta=0.5,
        )],
        base_station_info=BaseStationInfo(geo=BaseStationInfoGeo(dt=DATETIME, actual=True), uptime_s=21),
        base_station_additional_info=BaseStationAdditionalInfo(id=15015),
        base_station_logs=[BaseStationLogs(type=BaseStationLogsTypeEnum.info, dt=DATETIME, text='')],
    ).to_integration_meta(base_station_api_user_id=UUID('bca85c53-3819-4470-ad53-b2cba9cce603'), packet=DEVICE_PACKAGE_UNBP)) is IntegrationV0MessageMeta


def test_parsing_of_wrong_device_packet() -> None:
    assert HttpV0NeroBsPacket.parse(DEVICE_PACKET_NBFI) != HttpV0NeroBsPacket(
        data=[
            DevicePackage(
                dt=DATETIME,
                raw='FFFFFFFFFFFFFFFF',
                type=DevicePackageNetworkTypeEnum.uplink,
                protocol=DevicePackageProtocolTypeEnum.nbfi,
                data_nbfi=DataNbfi(mac=64, f_ask=True, iterator=3, multi=False, system=True, payload='FFFFFFFFFFFFFFFF'),
                sdr_info=SdrInfo(freq=870282100, freq_channel=871282100, sdr=0, baud_rate=50, rssi=-75, snr=65),
            ),
        ],
        base_station_environment_info=[BaseStationEnvironmentInfo(
            dt=datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
            spectrum=[0, 0, 0, 0, 0, 0],
            sdr=25,
            freq_carrier=0.1,
            freq_delta=0.5,
        )],
        base_station_info=BaseStationInfo(geo=BaseStationInfoGeo(dt=DATETIME, actual=True), uptime_s=21),
        base_station_additional_info=BaseStationAdditionalInfo(id=15015),
        base_station_logs=[BaseStationLogs(type=BaseStationLogsTypeEnum.info, dt=DATETIME, text='')],
    )
    assert HttpV0NeroBsPacket.parse(DEVICE_PACKET_UNBP) != HttpV0NeroBsPacket(
        data=[
            DevicePackage(
                dt=DATETIME,
                raw='FFFFFFFFFFFFFFFF',
                type=DevicePackageNetworkTypeEnum.downlink,
                protocol=DevicePackageProtocolTypeEnum.unbp,
                data_unbp=DataUnbp(mac=64, iterator=3, payload='FFFFFFFF', ack=True),
                sdr_info=SdrInfo(freq=871282100, freq_channel=871282100, sdr=0, baud_rate=50, rssi=-75, snr=65),
            ),
        ],
        base_station_environment_info=[BaseStationEnvironmentInfo(
            dt=datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
            spectrum=[0, 0, 0, 0, 0, 0],
            sdr=25,
            freq_carrier=0.1,
            freq_delta=0.5,
        )],
        base_station_info=BaseStationInfo(geo=BaseStationInfoGeo(dt=DATETIME, actual=True), uptime_s=21),
        base_station_additional_info=BaseStationAdditionalInfo(id=15015),
        base_station_logs=[BaseStationLogs(type=BaseStationLogsTypeEnum.info, dt=DATETIME, text='')],
    )
    assert HttpV0NeroBsPacket.parse(DEVICE_PACKET_UNBP2) != HttpV0NeroBsPacket(
        data=[
            DevicePackage(
                dt=DATETIME,
                raw='FFFFFFFFFFFFFFFF',
                type=DevicePackageNetworkTypeEnum.downlink,
                protocol=DevicePackageProtocolTypeEnum.unbp2,
                data_unbp2=DataUnbp2(mac=64, payload='FFFFFFFF'),
                sdr_info=SdrInfo(freq=871282100, freq_channel=871282100, sdr=0, baud_rate=50, rssi=-75, snr=65),
            ),
        ],
        base_station_environment_info=[BaseStationEnvironmentInfo(
            dt=datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
            spectrum=[0, 0, 0, 0, 0, 0],
            sdr=25,
            freq_carrier=0.1,
            freq_delta=0.5,
        )],
        base_station_info=BaseStationInfo(geo=BaseStationInfoGeo(dt=DATETIME, actual=True), uptime_s=21),
        base_station_additional_info=BaseStationAdditionalInfo(id=15015),
        base_station_logs=[BaseStationLogs(type=BaseStationLogsTypeEnum.info, dt=DATETIME, text='')],
    )


def test_check_validation_of_device_packet() -> None:
    try:
        HttpV0NeroBsPacket(
            data=[
                DevicePackage(
                    dt=DATETIME,
                    raw='FF',
                    type=DevicePackageNetworkTypeEnum.uplink,
                    protocol=DevicePackageProtocolTypeEnum.nbfi,
                    data_nbfi=DataNbfi(mac=0, f_ask=True, iterator=-1, multi=False, system=True, payload='FFFFFFFFFFFFFFF'),
                    sdr_info=SdrInfo(freq=0, freq_channel=-1, sdr=0, baud_rate=-1, rssi=-75, snr=65),
                ),
            ],
            base_station_environment_info=[BaseStationEnvironmentInfo(
                dt=datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
                spectrum=[0, 0, 0, 0, 0, 0],
                sdr=25,
                freq_carrier=0.1,
                freq_delta=0.5,
            )],
            base_station_info=BaseStationInfo(geo=BaseStationInfoGeo(dt=DATETIME, actual=True), uptime_s=-1),
            base_station_additional_info=BaseStationAdditionalInfo(id=15015),
            base_station_logs=[BaseStationLogs(type=BaseStationLogsTypeEnum.info, dt=DATETIME, text='')],
        )
    except ValidationError as exc:
        assert type(exc) is ValidationError

    try:
        HttpV0NeroBsPacket(
            data=[
                DevicePackage(
                    dt=DATETIME,
                    raw='FF',
                    type=DevicePackageNetworkTypeEnum.downlink,
                    protocol=DevicePackageProtocolTypeEnum.unbp,
                    data_unbp=DataUnbp(mac=0, iterator=-1, payload='FF', ack=True),
                    sdr_info=SdrInfo(freq=0, freq_channel=-1, sdr=0, baud_rate=-1, rssi=-75, snr=65),
                ),
            ],
            base_station_environment_info=[BaseStationEnvironmentInfo(
                dt=datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
                spectrum=[0, 0, 0, 0, 0, 0],
                sdr=25,
                freq_carrier=0.1,
                freq_delta=0.5,
            )],
            base_station_info=BaseStationInfo(geo=BaseStationInfoGeo(dt=DATETIME, actual=True), uptime_s=-1),
            base_station_additional_info=BaseStationAdditionalInfo(id=15015),
            base_station_logs=[BaseStationLogs(type=BaseStationLogsTypeEnum.info, dt=DATETIME, text='')],
        )
    except ValidationError as exc:
        assert type(exc) is ValidationError

    try:
        HttpV0NeroBsPacket(
            data=[
                DevicePackage(
                    dt=DATETIME,
                    raw='FF',
                    type=DevicePackageNetworkTypeEnum.downlink,
                    protocol=DevicePackageProtocolTypeEnum.unbp2,
                    data_unbp2=DataUnbp2(mac=0, payload='FF'),
                    sdr_info=SdrInfo(freq=0, freq_channel=-1, sdr=0, baud_rate=-1, rssi=-75, snr=65),
                ),
            ],
            base_station_environment_info=[BaseStationEnvironmentInfo(
                dt=datetime(year=2022, month=10, day=15, hour=4, minute=16, second=11),
                spectrum=[0, 0, 0, 0, 0, 0],
                sdr=25,
                freq_carrier=0.1,
                freq_delta=0.5,
            )],
            base_station_info=BaseStationInfo(geo=BaseStationInfoGeo(dt=DATETIME, actual=True), uptime_s=-1),
            base_station_additional_info=BaseStationAdditionalInfo(id=15015),
            base_station_logs=[BaseStationLogs(type=BaseStationLogsTypeEnum.info, dt=DATETIME, text='')],
        )
    except ValidationError as exc:
        assert type(exc) is ValidationError

    try:
        HttpV0NeroBsPacket.parse(DEVICE_PACKET_WRONG)
    except DataGatewayBSProtocolParsingError as exc:
        assert type(exc) is DataGatewayBSProtocolParsingError



from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_08b_counter_volume import \
    SmpmUlDeviceJupiter08BCounterVolumeData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_jupiter_08b_counter_volume() -> None:
    case_serialized = bytes.fromhex("6d000000040000e0")
    assert SmpmUlDeviceJupiter08BCounterVolumeData(volume_channel_1=67108.864, volume_channel_2=67108.864, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterVolumeData.serialize(SmpmUlDeviceJupiter08BCounterVolumeData(volume_channel_1=67108.864, volume_channel_2=67108.864, event_reset=True, event_low_battery_level=True))  # noqa: E501
    case_serialized = bytes.fromhex("6d000000000000c0")  # overflow
    assert SmpmUlDeviceJupiter08BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterVolumeData.serialize(SmpmUlDeviceJupiter08BCounterVolumeData(event_low_battery_level=True, event_reset=True, volume_channel_1=0.0, volume_channel_2=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("6d000000000000c0")  # overflow
    assert SmpmUlDeviceJupiter08BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterVolumeData.serialize(SmpmUlDeviceJupiter08BCounterVolumeData(event_low_battery_level=True, event_reset=True, volume_channel_1=0.0, volume_channel_2=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("6d000000000000c0")  # overflow
    assert SmpmUlDeviceJupiter08BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterVolumeData.serialize(SmpmUlDeviceJupiter08BCounterVolumeData(event_low_battery_level=True, event_reset=True, volume_channel_1=268435.456, volume_channel_2=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("6d000000000000c0")  # overflow
    assert SmpmUlDeviceJupiter08BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterVolumeData.serialize(SmpmUlDeviceJupiter08BCounterVolumeData(event_low_battery_level=True, event_reset=True, volume_channel_1=268435.456, volume_channel_2=268435.456))  # noqa: E501



import zlib
from typing import NamedTuple, Dict, Any, List

from data_gateway_sdk.errors import DataGatewayDeviceProtocolParsingError
from data_gateway_sdk.utils.buf_ref import BufRef


# PACKET (N/A bits)   unbp_uplink_len
#
# RESULT int:        230523608866647489049872409590682171
# RESULT bin:  MSB   00101100 01100101 10110001 00000001 00010110 00111111 01111111 00000011 00000000 00000000 00000000 00000000 00100010 11010010 00111011   LSB
# RESULT hex:  LE    3b d2 22 00 00 00 00 03 7f 3f 16 01 b1 65 2c
#
# name            type  size  value(int)                                                                                                                data(bits)
# --------------  ----  ----  ----------  ------------------------------------------------------------------------------------------------------------------------
# identifier      u16     16       53819                                                                                                          1101001000111011
# mac             u32     32          34                                                                          00000000000000000000000000100010
# RESERVED        u3       3           0                                                                       000
# message_id      u5       5           0                                                                  00000
# payload.LEN     u8       8           3                                                          00000011
# payload.0.byte  u8       8         127                                                  01111111
# payload.1.byte  u8       8          63                                          00111111
# payload.2.byte  u8       8          22                                  00010110
# crc             u32     32   744861953  00101100011001011011000100000001


class UnbpUplinkLenData(NamedTuple):
    mac: int
    message_id: int
    payload: List[int]
    crc: int

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((53819) & (2 ** (16) - 1)) << size
        size += 16
        assert isinstance(data.mac, int)
        assert 0 <= data.mac <= 4294967295
        result |= ((data.mac) & (2 ** (32) - 1)) << size
        size += 32
        result |= ((0) & (2 ** (3) - 1)) << size
        size += 3
        assert isinstance(data.message_id, int)
        assert 0 <= data.message_id <= 31
        result |= ((data.message_id) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.payload, list)
        assert isinstance(len(data.payload), int)
        assert 0 <= len(data.payload) <= 255
        result |= ((len(data.payload)) & (2 ** (8) - 1)) << size
        size += 8
        for payload__item_tmp1 in data.payload[:255]:
            assert isinstance(payload__item_tmp1, int)
            assert 0 <= payload__item_tmp1 <= 255
            result |= ((payload__item_tmp1) & (2 ** (8) - 1)) << size
            size += 8
        assert isinstance(data.crc, int)
        assert 0 <= data.crc <= 4294967295
        result |= ((data.crc) & (2 ** (32) - 1)) << size
        size += 32
        return result.to_bytes(size // 8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'UnbpUplinkLenData':
        result__el_tmp2: Dict[str, Any] = dict()
        prev_ends_at = buf.ends_at

        try:
            if 53819 != buf.shift(16):
                raise DataGatewayDeviceProtocolParsingError("identifier: buffer doesn't match value")
        except ValueError:
            raise DataGatewayDeviceProtocolParsingError("identifier: buffer doesn't match value")

        try:
            result__el_tmp2["mac"] = buf.shift(32) + 0
        except ValueError:
            raise DataGatewayDeviceProtocolParsingError("mac: buffer doesn't match value")

        try:
            if 0 != buf.shift(3):
                raise DataGatewayDeviceProtocolParsingError("RESERVED: buffer doesn't match value")
        except ValueError:
            raise DataGatewayDeviceProtocolParsingError("RESERVED: buffer doesn't match value")

        try:
            result__el_tmp2["message_id"] = buf.shift(5) + 0
        except ValueError:
            raise DataGatewayDeviceProtocolParsingError("message_id: buffer doesn't match value")
        payload__res_tmp5: List[int] = []

        try:
            length_tmp6 = buf.shift(8) + 0
            for _payload__i_tmp3 in range(length_tmp6):
                payload__item_tmp4 = buf.shift(8) + 0
                payload__res_tmp5.append(payload__item_tmp4)
            result__el_tmp2["payload"] = payload__res_tmp5
        except ValueError:
            raise DataGatewayDeviceProtocolParsingError("payload: buffer doesn't match value")

        try:
            result__el_tmp2["crc"] = buf.shift(32) + 0
            if result__el_tmp2["crc"] in [0, 4294967295]:
                raise DataGatewayDeviceProtocolParsingError("CRC: CRC invalid ")
        except ValueError:
            raise DataGatewayDeviceProtocolParsingError("CRC: buffer doesn't match value")
        else:
            size = 8 + len(result__el_tmp2["payload"])
            array_without_crc: bytes = buf.get_bits(size=size * 8, start=-(buf.ends_at - prev_ends_at)).to_bytes(byteorder="big", length=size)
            crc = zlib.crc32(array_without_crc[::-1])
            if result__el_tmp2["crc"] != crc:
                raise DataGatewayDeviceProtocolParsingError("CRC: CRC sum mismatched ")
            result = UnbpUplinkLenData(**result__el_tmp2)
        return result



from typing import Dict, List, Any
from datetime import datetime, tzinfo

from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent
from data_aggregator_sdk.integration_message import IntegrationV0MessageData

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   water5_device_warm_reset_v_metano_a
#
# RESULT int:        9795
# RESULT bin:  MSB   00000000 00000000 00000000 00000000 00000000 00000000 00100110 01000011   LSB
# RESULT hex:  LE    43 26 00 00 00 00 00 00
#
# name              type  size  value(int)                                                        data(bits)
# ----------------  ----  ----  ----------  ----------------------------------------------------------------
# pack_id           u8       8          67                                                          01000011
# UNUSED            -        1           0                                                         0
# watchdog_reset    bool     1           1                                                        1
# low_voltage       bool     1           1                                                       1
# software_reset    bool     1           0                                                      0
# UNUSED            -        1           0                                                     0
# pin_reset         bool     1           1                                                    1
# UNUSED            -        1           0                                                   0
# hard_fault_error  bool     1           0                                                  0
# RESERVED          u48     48           0  000000000000000000000000000000000000000000000000


class Water5DeviceWarmResetVMetanoAData(Packet):
    watchdog_reset: bool
    low_voltage: bool
    software_reset: bool
    pin_reset: bool
    hard_fault_error: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((67) & (2 ** (8) - 1)) << size
        size += 8
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.watchdog_reset, bool)
        result |= ((int(data.watchdog_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.low_voltage, bool)
        result |= ((int(data.low_voltage)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.software_reset, bool)
        result |= ((int(data.software_reset)) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.pin_reset, bool)
        result |= ((int(data.pin_reset)) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hard_fault_error, bool)
        result |= ((int(data.hard_fault_error)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'Water5DeviceWarmResetVMetanoAData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 67 != buf.shift(8):
            raise ValueError("pack_id: buffer doesn't match value")
        buf.shift(1)
        result__el_tmp1["watchdog_reset"] = bool(buf.shift(1))
        result__el_tmp1["low_voltage"] = bool(buf.shift(1))
        result__el_tmp1["software_reset"] = bool(buf.shift(1))
        buf.shift(1)
        result__el_tmp1["pin_reset"] = bool(buf.shift(1))
        buf.shift(1)
        result__el_tmp1["hard_fault_error"] = bool(buf.shift(1))
        buf.shift(48)
        result = Water5DeviceWarmResetVMetanoAData(**result__el_tmp1)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                events=[
                    *([] if any((self.pin_reset, self.low_voltage, self.software_reset, self.watchdog_reset)) else [IntegrationV0MessageEvent.RESET]),
                    *([] if not self.pin_reset else [IntegrationV0MessageEvent.RESET_PIN]),
                    *([] if not self.low_voltage else [IntegrationV0MessageEvent.RESET_LOW_VOLTAGE]),
                    *([] if not self.software_reset else [IntegrationV0MessageEvent.RESET_SOFTWARE]),
                    *([] if not self.watchdog_reset else [IntegrationV0MessageEvent.RESET_WATCHDOG]),
                    *([] if not self.hard_fault_error else [IntegrationV0MessageEvent.RESET_HARD_FAULT]),
                ],
            ),
        ]



from data_gateway_sdk.protocols.water5_v_nero.water5_device_weekly_ch1_impulse_v2 import \
    Water5DeviceWeeklyCh1ImpulseV2Data
from data_gateway_sdk.utils.buf_ref import BufRef


def test_water5_device_weekly_ch1_impulse_v2() -> None:
    case = Water5DeviceWeeklyCh1ImpulseV2Data(value=12313, battery_fract=67, battery_int=3)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=0, battery_int=2, value=0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=0, battery_int=2, value=12313)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=0, battery_int=2, value=16777215)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=0, battery_int=3, value=0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=0, battery_int=3, value=12313)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=0, battery_int=3, value=16777215)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=67, battery_int=2, value=0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=67, battery_int=2, value=12313)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=67, battery_int=2, value=16777215)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=67, battery_int=3, value=0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=67, battery_int=3, value=12313)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=67, battery_int=3, value=16777215)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=127, battery_int=2, value=0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=127, battery_int=2, value=12313)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=127, battery_int=2, value=16777215)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=127, battery_int=3, value=0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=127, battery_int=3, value=12313)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1ImpulseV2Data(battery_fract=127, battery_int=3, value=16777215)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1ImpulseV2Data.parse(BufRef(case.serialize()))



from datetime import datetime, tzinfo
from typing import Tuple, TypeVar, Type, List, Any

from data_aggregator_sdk.integration_message import IntegrationV0MessageData
from pydantic import BaseModel

from data_gateway_sdk.utils.buf_ref import BufRef

T = TypeVar('T', bound='Packet')


class Packet(BaseModel):
    @classmethod
    def parse(cls: Type[T], buf: BufRef) -> T:
        raise NotImplementedError()

    def serialize(self) -> bytes:
        raise NotImplementedError()

    def to_integration_data(self, *, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        raise NotImplementedError()

    @staticmethod
    def parse_packets(payload: bytes, pack_types: Tuple[Type[T], ...]) -> List[T]:
        result: List[T] = []
        while True:
            prev_res_len = len(result)
            for pack_type in pack_types:
                try:
                    buf = BufRef(payload)
                    result.append(pack_type.parse(buf))
                    payload = buf.get_last_bytes()
                    break
                except ValueError:
                    continue
            if prev_res_len == len(result):
                return []
            if len(payload) == 0:
                break
            if all(not i for i in payload):
                break
        return result



from datetime import datetime, timedelta, time, tzinfo

from dateutil.relativedelta import relativedelta
from pytz import timezone


def days_ago_calculation(received_at: datetime, device_tz: tzinfo, days_ago_time: time, days_ago: timedelta | relativedelta) -> datetime:
    return (datetime.combine(
        (received_at.astimezone(device_tz) - days_ago).date(),
        days_ago_time,
        device_tz,
    ) + timedelta(days=1)).astimezone(timezone('UTC')) if days_ago else received_at



from decimal import Decimal
from datetime import timedelta, datetime, tzinfo
from typing import List, Any, Dict
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, \
    IntegrationV0MessageRelay, IntegrationV0MessageClock, IntegrationV0MessageSensor
from data_aggregator_sdk.constants.enums import SensorType
from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_energy_16b_info
#
# RESULT int:        38094212464183996
# RESULT bin:  MSB   00000000000000000000000000000000000000000000000000000000000000000000000010000111010101100111110000000001111011010000101010111100   LSB                             # noqa: E501
# RESULT hex:  LE    bc 0a ed 01 7c 56 87 00 00 00 00 00 00 00 00 00
#
#
# name                  type       size  value(int)                                                                                                                        data(bits)   # noqa: E501
# --------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------   # noqa: E501
# packet_type_id.0.VAL  u7            7          60                                                                                                                           0111100   # noqa: E501
# packet_type_id.0.DFF  bool          1           1                                                                                                                          1          # noqa: E501
# packet_type_id.1.VAL  u2            2           2                                                                                                                        10           # noqa: E501
# packet_type_id.1.DFF  bool          1           0                                                                                                                       0             # noqa: E501
# battery_volts         uf7p1         7          33                                                                                                                0100001              # noqa: E501
# temperature           u8            8         123                                                                                                        01111011                     # noqa: E501
# datetime              timedelta    31   567648000                                                                         0100001110101011001111100000000                             # noqa: E501
# relay_is_active       bool          1           0                                                                        0                                                            # noqa: E501
# RESERVED              u70          70           0  0000000000000000000000000000000000000000000000000000000000000000000000                                                             # noqa: E501


class SmpmUlDeviceEnergy16BInfoData(Packet):
    battery_volts: float
    temperature: int
    date_time: timedelta
    relay_is_active: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((60) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((2) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 12.7
        result |= ((int(round(float(data.battery_volts) * 10.0, 0))) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.temperature, int)
        assert -100 <= data.temperature <= 155
        result |= (((data.temperature + 100)) & (2 ** (8) - 1)) << size
        size += 8
        isinstance(data.date_time, (int, timedelta))
        datetime_tmp1 = int(data.date_time.total_seconds() // 1 if isinstance(data.date_time, timedelta) else data.date_time // 1)
        assert 0 <= datetime_tmp1 <= 2147483647
        result |= ((datetime_tmp1) & (2 ** (31) - 1)) << size
        size += 31
        assert isinstance(data.relay_is_active, bool)
        result |= ((int(data.relay_is_active)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BInfoData':
        result__el_tmp2: Dict[str, Any] = dict()
        if 60 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 2 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp2["battery_volts"] = round(buf.shift(7) / 10.0, 1)
        result__el_tmp2["temperature"] = buf.shift(8) + -100
        result__el_tmp2["date_time"] = timedelta(seconds=buf.shift(31) * 1)
        result__el_tmp2["relay_is_active"] = bool(buf.shift(1))
        result = SmpmUlDeviceEnergy16BInfoData(**result__el_tmp2)
        buf.shift(70)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
                relay=[
                    IntegrationV0MessageRelay(
                        value=self.relay_is_active,
                    ),
                ],
                clock=[
                    IntegrationV0MessageClock(
                        value=datetime(year=2020, month=1, day=1) + self.date_time,
                    ),
                ],
            ),
        ]



from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_08b_counter_ch2 import SmpmUlDeviceJupiter08BCounterCh2Data
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_jupiter_08b_counter_ch2() -> None:
    case_serialized = bytes.fromhex("67ffffff7fa16e00")
    assert SmpmUlDeviceJupiter08BCounterCh2Data(value=2147483647, battery_volts=3.3, temperature=23, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterCh2Data.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterCh2Data.serialize(SmpmUlDeviceJupiter08BCounterCh2Data(value=2147483647, battery_volts=3.3, temperature=23, event_reset=True, event_low_battery_level=True))  # noqa: E501
    case_serialized = bytes.fromhex("6700000000c07f00")  # overflow
    assert SmpmUlDeviceJupiter08BCounterCh2Data(value=0, battery_volts=0.0, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterCh2Data.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterCh2Data.serialize(SmpmUlDeviceJupiter08BCounterCh2Data(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=92, value=0))  # noqa: E501
    case_serialized = bytes.fromhex("6700000000c07f00")  # overflow
    assert SmpmUlDeviceJupiter08BCounterCh2Data(value=0, battery_volts=0.0, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterCh2Data.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterCh2Data.serialize(SmpmUlDeviceJupiter08BCounterCh2Data(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=92, value=8589934592))  # noqa: E501
    case_serialized = bytes.fromhex("6700000000006000")  # overflow
    assert SmpmUlDeviceJupiter08BCounterCh2Data(value=0, battery_volts=0.0, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterCh2Data.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterCh2Data.serialize(SmpmUlDeviceJupiter08BCounterCh2Data(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=-35, value=0))  # noqa: E501
    case_serialized = bytes.fromhex("6700000000006000")  # overflow
    assert SmpmUlDeviceJupiter08BCounterCh2Data(value=0, battery_volts=0.0, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterCh2Data.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterCh2Data.serialize(SmpmUlDeviceJupiter08BCounterCh2Data(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=-35, value=8589934592))  # noqa: E501
    case_serialized = bytes.fromhex("6700000000ff7f00")  # overflow
    assert SmpmUlDeviceJupiter08BCounterCh2Data(value=0, battery_volts=6.3, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterCh2Data.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterCh2Data.serialize(SmpmUlDeviceJupiter08BCounterCh2Data(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=92, value=0))  # noqa: E501
    case_serialized = bytes.fromhex("6700000000ff7f00")  # overflow
    assert SmpmUlDeviceJupiter08BCounterCh2Data(value=0, battery_volts=6.3, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterCh2Data.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterCh2Data.serialize(SmpmUlDeviceJupiter08BCounterCh2Data(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=92, value=8589934592))  # noqa: E501
    case_serialized = bytes.fromhex("67000000003f6000")  # overflow
    assert SmpmUlDeviceJupiter08BCounterCh2Data(value=0, battery_volts=6.3, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterCh2Data.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterCh2Data.serialize(SmpmUlDeviceJupiter08BCounterCh2Data(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=-35, value=0))  # noqa: E501
    case_serialized = bytes.fromhex("67000000003f6000")  # overflow
    assert SmpmUlDeviceJupiter08BCounterCh2Data(value=0, battery_volts=6.3, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterCh2Data.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterCh2Data.serialize(SmpmUlDeviceJupiter08BCounterCh2Data(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=-35, value=8589934592))  # noqa: E501



from datetime import datetime, tzinfo
from decimal import Decimal
from typing import Any, List

from data_aggregator_sdk.constants.enums import SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageConsumption, CounterType, ResourceType, \
    IntegrationV0MessageSensor
from pydantic import BaseModel
from pysmp import decode_getset as smp_getset  # type: ignore

from data_gateway_sdk.errors import DataGatewayDeviceProtocolParsingError
from data_gateway_sdk.protocols.device_packet.device_packet import DevicePacket


class SmpDaily(BaseModel):
    info_id_code: int
    sdata_count_ch1_code: int
    sdata_count_ch2_code: int
    sdata_temperature_code: int
    sdata_battery_code: float
    sdata_datetime_code: datetime


class SmpV0DevicePacket(DevicePacket):
    packet: SmpDaily

    @classmethod
    def parse(cls, payload: bytes, **kwargs: Any) -> List['SmpV0DevicePacket']:
        try:
            return [SmpV0DevicePacket(
                packet=SmpDaily(
                    **{cmd.name.replace('SMP_', '').lower(): (val.value.isoformat() if isinstance(val.value, datetime) else val.value) for cmd, val in smp_getset(payload)},  # type: ignore
                ),
            )]
        except Exception as e:  # noqa: B902
            raise DataGatewayDeviceProtocolParsingError(f'invalid payload {payload.hex()}', e)

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.packet.sdata_battery_code)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.packet.sdata_temperature_code)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(self.packet.sdata_count_ch1_code)),
                        overloading_value=None,
                    ),
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=2,
                        value=Decimal(str(self.packet.sdata_count_ch2_code)),
                        overloading_value=None,
                    ),
                ],
            ),
        ]



from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_04b_event import SmpmUlDeviceWaterMeter04BEventData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_water_meter_04b_event() -> None:
    case_serialized = bytes.fromhex("0ce13a06")
    assert SmpmUlDeviceWaterMeter04BEventData(battery_volts=3.3, event_low_battery_level=True, event_temperature_limits=True, temperature=23, event_case_was_opened=False, event_magnet=False, event_reset=True, event_sensor_error=True) == SmpmUlDeviceWaterMeter04BEventData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter04BEventData.serialize(SmpmUlDeviceWaterMeter04BEventData(battery_volts=3.3, event_low_battery_level=True, event_temperature_limits=True, temperature=23, event_case_was_opened=False, event_magnet=False, event_reset=True, event_sensor_error=True))  # noqa: E501
    case_serialized = bytes.fromhex("0cc0ff07")  # overflow
    assert SmpmUlDeviceWaterMeter04BEventData(battery_volts=0.0, event_low_battery_level=True, event_temperature_limits=True, temperature=92, event_case_was_opened=True, event_magnet=True, event_reset=True, event_sensor_error=True) == SmpmUlDeviceWaterMeter04BEventData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter04BEventData.serialize(SmpmUlDeviceWaterMeter04BEventData(battery_volts=0.0, event_case_was_opened=True, event_low_battery_level=True, event_magnet=True, event_reset=True, event_sensor_error=True, event_temperature_limits=True, temperature=92))  # noqa: E501
    case_serialized = bytes.fromhex("0cc08007")  # overflow
    assert SmpmUlDeviceWaterMeter04BEventData(battery_volts=0.0, event_low_battery_level=True, event_temperature_limits=True, temperature=-35, event_case_was_opened=True, event_magnet=True, event_reset=True, event_sensor_error=True) == SmpmUlDeviceWaterMeter04BEventData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter04BEventData.serialize(SmpmUlDeviceWaterMeter04BEventData(battery_volts=0.0, event_case_was_opened=True, event_low_battery_level=True, event_magnet=True, event_reset=True, event_sensor_error=True, event_temperature_limits=True, temperature=-35))  # noqa: E501
    case_serialized = bytes.fromhex("0cffff07")  # overflow
    assert SmpmUlDeviceWaterMeter04BEventData(battery_volts=6.3, event_low_battery_level=True, event_temperature_limits=True, temperature=92, event_case_was_opened=True, event_magnet=True, event_reset=True, event_sensor_error=True) == SmpmUlDeviceWaterMeter04BEventData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter04BEventData.serialize(SmpmUlDeviceWaterMeter04BEventData(battery_volts=6.3, event_case_was_opened=True, event_low_battery_level=True, event_magnet=True, event_reset=True, event_sensor_error=True, event_temperature_limits=True, temperature=92))  # noqa: E501
    case_serialized = bytes.fromhex("0cff8007")  # overflow
    assert SmpmUlDeviceWaterMeter04BEventData(battery_volts=6.3, event_low_battery_level=True, event_temperature_limits=True, temperature=-35, event_case_was_opened=True, event_magnet=True, event_reset=True, event_sensor_error=True) == SmpmUlDeviceWaterMeter04BEventData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter04BEventData.serialize(SmpmUlDeviceWaterMeter04BEventData(battery_volts=6.3, event_case_was_opened=True, event_low_battery_level=True, event_magnet=True, event_reset=True, event_sensor_error=True, event_temperature_limits=True, temperature=-35))  # noqa: E501



from data_gateway_sdk.protocols.water5_v_nero.water5_device_warm_reset_v_fluo_a import Water5DeviceWarmResetVFluoAData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_water5_device_warm_reset_v_fluo_a() -> None:
    case = Water5DeviceWarmResetVFluoAData(pin_reset=True, low_voltage=True, software_reset=False, watchdog_reset=True)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=False, pin_reset=False, software_reset=False, watchdog_reset=False)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=False, pin_reset=False, software_reset=False, watchdog_reset=True)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=False, pin_reset=False, software_reset=True, watchdog_reset=False)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=False, pin_reset=False, software_reset=True, watchdog_reset=True)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=False, pin_reset=True, software_reset=False, watchdog_reset=False)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=False, pin_reset=True, software_reset=False, watchdog_reset=True)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=False, pin_reset=True, software_reset=True, watchdog_reset=False)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=False, pin_reset=True, software_reset=True, watchdog_reset=True)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=True, pin_reset=False, software_reset=False, watchdog_reset=False)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=True, pin_reset=False, software_reset=False, watchdog_reset=True)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=True, pin_reset=False, software_reset=True, watchdog_reset=False)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=True, pin_reset=False, software_reset=True, watchdog_reset=True)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=True, pin_reset=True, software_reset=False, watchdog_reset=False)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=True, pin_reset=True, software_reset=False, watchdog_reset=True)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=True, pin_reset=True, software_reset=True, watchdog_reset=False)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
    case = Water5DeviceWarmResetVFluoAData(low_voltage=True, pin_reset=True, software_reset=True, watchdog_reset=True)  # noqa: E501
    assert case == Water5DeviceWarmResetVFluoAData.parse(BufRef(case.serialize()))
